#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os.path
import re
from pkgutil import walk_packages

from setuptools import setup


def find_packages(path=["."], prefix=""):
    yield prefix
    prefix = prefix + "."
    for _, name, ispkg in walk_packages(path, prefix):
        if ispkg:
            yield name


with open(os.path.join("cumulusci", "version.txt"), "r") as version_file:
    version = version_file.read().strip()

with open("README.rst", "rb") as readme_file:
    readme = readme_file.read().decode("utf-8")

with open("HISTORY.rst", "rb") as history_file:
    history = history_file.read().decode("utf-8")

with open("requirements/prod.txt") as requirements_file:
    requirements = []
    for req in requirements_file.read().splitlines():
        # skip comments and hash lines
        if re.match(r"\s*#", req) or re.match(r"\s*--hash", req):
            continue
        else:
            req = req.split(" ")[0]
            requirements.append(req)

setup(
    name="cumulusci",
    version=version,
    description="Build and release tools for Salesforce developers",
    long_description=readme + u"\n\n" + history,
    long_description_content_type="text/x-rst",
    author="Salesforce.org",
    author_email="jlantz@salesforce.com",
    url="https://github.com/SFDO-Tooling/CumulusCI",
    packages=list(find_packages(["cumulusci"], "cumulusci")),
    package_dir={"cumulusci": "cumulusci"},
    entry_points={
        "console_scripts": [
            "cci=cumulusci.cli.cci:main",
            "snowfakery=snowfakery.cli:main",
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="BSD license",
    zip_safe=False,
    keywords="cumulusci",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    test_suite="cumulusci.core.tests",
    python_requires=">=3.6",
)



import pytest
import responses

from cumulusci.core.config import ServiceConfig
from cumulusci.core.config.marketing_cloud_service_config import (
    MarketingCloudServiceConfig,
)
from cumulusci.tasks.marketing_cloud.tests import test_api_soap_envelopes as envelopes
from cumulusci.tests.util import create_project_config

from ..api import CreateSubscriberAttribute, CreateUser, UpdateUserRole


@pytest.fixture
def project_config():
    project_config = create_project_config()
    project_config.keychain.set_service(
        "oauth2_client",
        "test",
        ServiceConfig(
            {
                "client_id": "MC_CLIENT_ID",
                "client_secret": "BOGUS",
                "auth_uri": "https://TSSD.auth.marketingcloudapis.com/v2/authorize",
                "token_uri": "https://TSSD.auth.marketingcloudapis.com/v2/token",
                "callback_url": "https://127.0.0.1:8080/",
            },
            "test",
            project_config.keychain,
        ),
        False,
    )
    project_config.keychain.set_service(
        "marketing_cloud",
        "test",
        MarketingCloudServiceConfig(
            {
                "oauth2_client": "test",
                "refresh_token": "REFRESH",
                "soap_instance_url": "https://TSSD.soap.marketingcloudapis.com/",
            },
            "test",
            project_config.keychain,
        ),
        False,
    )
    return project_config


@responses.activate
def test_marketing_cloud_create_subscriber_attribute_task(create_task, project_config):
    responses.add(
        "POST",
        "https://tssd.auth.marketingcloudapis.com/v2/token",
        json={"access_token": "ACCESS_TOKEN"},
    )
    responses.add(
        "POST",
        "https://tssd.soap.marketingcloudapis.com/Service.asmx",
        body=envelopes.CREATE_SUBSCRIBER_ATTRIBUTE_EXPECTED_SOAP_RESPONSE,
    )

    task = create_task(
        CreateSubscriberAttribute,
        {
            "attribute_name": "Test Subscriber Attribute",
        },
        project_config=project_config,
    )
    task()
    assert task.return_values == {"success": True}


@responses.activate
def test_marketing_cloud_create_user_task(create_task, project_config):
    responses.add(
        "POST",
        "https://tssd.auth.marketingcloudapis.com/v2/token",
        json={"access_token": "ACCESS_TOKEN"},
    )
    responses.add(
        "POST",
        "https://tssd.soap.marketingcloudapis.com/Service.asmx",
        body=envelopes.CREATE_USER_EXPECTED_SOAP_RESPONSE,
    )

    task = create_task(
        CreateUser,
        {
            "parent_bu_mid": "523005197",
            "default_bu_mid": "523008403",
            "external_key": "Don_Draper_Key_1926",
            "user_name": "Don Draper",
            "user_email": "don.draper@sterlingcooper.com",
            "user_password": "SterlingCooperDraperPryce1!",
            "user_username": "sterling-don",
            "role_id": "31",
        },
        project_config=project_config,
    )
    task()
    assert task.return_values == {"success": True}


@responses.activate
def test_marketing_cloud_update_user_role_task(create_task, project_config):
    responses.add(
        "POST",
        "https://tssd.auth.marketingcloudapis.com/v2/token",
        json={"access_token": "ACCESS_TOKEN"},
    )
    responses.add(
        "POST",
        "https://tssd.soap.marketingcloudapis.com/Service.asmx",
        body=envelopes.UPDATE_USER_ROLE_EXPECTED_SOAP_RESPONSE,
    )

    task = create_task(
        UpdateUserRole,
        {
            "account_mid": "523005197",
            "external_key": "Don_Draper_Key_1926",
            "user_name": "Partner Don Draper",
            "user_email": "don.draper@sterlingcooper.com",
            "user_password": "SterlingCooperDraperPryce1!",
            "role_id": "31",
        },
        project_config=project_config,
    )
    task()
    assert task.return_values == {"success": True}



from cumulusci.core.tasks import BaseTask


class BaseMarketingCloudTask(BaseTask):
    """Base task for interacting with Marketing Cloud

    For API calls to a MC tenant, you can get a fresh
    access token via the marketing cloud config like so:

    self.mc_config.access_token
    """

    salesforce_task = False

    def _init_task(self):
        super()._init_task()
        self.mc_config = self.project_config.keychain.get_service("marketing_cloud")



""" Tasks are the basic unit of execution in CumulusCI.

Subclass BaseTask or a descendant to define custom task logic
"""
import contextlib
import logging
import os
import re
import threading
import time

from cumulusci import __version__
from cumulusci.core.debug import get_debug_mode
from cumulusci.core.exceptions import (
    ServiceNotConfigured,
    ServiceNotValid,
    TaskOptionsError,
    TaskRequiresSalesforceOrg,
)
from cumulusci.utils import cd
from cumulusci.utils.logging import redirect_output_to_logger

CURRENT_TASK = threading.local()

PROJECT_CONFIG_RE = re.compile(r"\$project_config.(\w+)")
CAPTURE_TASK_OUTPUT = os.environ.get("CAPTURE_TASK_OUTPUT")


# We can't use contextlib.nullcontext yet
# because it isn't present in Python 3.6
@contextlib.contextmanager
def nullcontext():
    yield


@contextlib.contextmanager
def stacked_task(task):
    if not hasattr(CURRENT_TASK, "stack"):
        CURRENT_TASK.stack = []
    CURRENT_TASK.stack.append(task)
    try:
        yield
    finally:
        CURRENT_TASK.stack.pop()


class BaseTask(object):
    """BaseTask provides the core execution logic for a Task

    Subclass BaseTask and provide a `_run_task()` method with your
    code.
    """

    task_docs = ""
    task_options = {}
    salesforce_task = False  # Does this task require a salesforce org?

    def __init__(
        self,
        project_config,
        task_config,
        org_config=None,
        flow=None,
        name=None,
        stepnum=None,
        logger=None,
        **kwargs,
    ):
        self.project_config = project_config
        self.task_config = task_config
        self.org_config = org_config
        self.logger = logger
        self._reset_poll()

        # dict of return_values that can be used by task callers
        self.return_values = {}

        # simple result object for introspection, often a return_code
        self.result = None

        # the flow for this task execution
        self.flow = flow

        # the tasks name in the flow
        self.name = name

        # the tasks stepnumber in the flow
        self.stepnum = stepnum

        self.debug_mode = get_debug_mode()

        if not self.logger:
            self._init_logger()

        self._init_options(kwargs)
        self._validate_options()

    def _init_logger(self):
        """Initializes self.logger"""
        if self.flow:
            self.logger = self.flow.logger.getChild(self.__class__.__name__)
        else:
            self.logger = logging.getLogger(__name__)

    def _init_options(self, kwargs):
        """Initializes self.options"""
        if self.task_config.options is None:
            self.options = {}
        else:
            self.options = self.task_config.options.copy()

        if kwargs:
            self.options.update(kwargs)

        # Handle dynamic lookup of project_config values via $project_config.attr
        for option, value in self.options.items():
            if isinstance(value, str):
                value = PROJECT_CONFIG_RE.sub(
                    lambda match: str(
                        getattr(self.project_config, match.group(1), None)
                    ),
                    value,
                )
                self.options[option] = value

    def _validate_options(self):
        missing_required = []
        for name, config in list(self.task_options.items()):
            if config.get("required") is True and name not in self.options:
                missing_required.append(name)

        if missing_required:
            required_opts = ",".join(missing_required)
            raise TaskOptionsError(
                f"{self.__class__.__name__} requires the options ({required_opts}) and no values were provided"
            )

    def _update_credentials(self):
        """Override to do any logic to refresh credentials"""
        pass

    def _init_task(self):
        """Override to implement dynamic logic for initializing the task."""
        pass

    def __call__(self):
        if self.salesforce_task and not self.org_config:
            raise TaskRequiresSalesforceOrg(
                "This task requires a salesforce org. "
                "Use org default <name> to set a default org "
                "or pass the org name with the --org option"
            )
        self._update_credentials()
        self._init_task()

        with stacked_task(self):
            self.working_path = os.getcwd()
            path = self.project_config.repo_root if self.project_config else None
            with cd(path):
                with (
                    redirect_output_to_logger(self.logger)
                    if CAPTURE_TASK_OUTPUT
                    else nullcontext()
                ):
                    self._log_begin()
                    self.result = self._run_task()
                    return self.return_values

    def _run_task(self):
        """Subclasses should override to provide their implementation"""
        raise NotImplementedError("Subclasses should provide their own implementation")

    def _log_begin(self):
        """Log the beginning of the task execution"""
        self.logger.info(f"Beginning task: {self.__class__.__name__}")
        if self.salesforce_task and not self.flow:
            self.logger.info(f"As user: {self.org_config.username}")
            self.logger.info(f"In org: {self.org_config.org_id}")
        self.logger.info("")

    def _retry(self):
        while True:
            try:
                self._try()
                break
            except Exception as e:
                if not (self.options["retries"] and self._is_retry_valid(e)):
                    raise
                if self.options["retry_interval"]:
                    self.logger.warning(
                        f"Sleeping for {self.options['retry_interval']} seconds before retry..."
                    )
                    time.sleep(self.options["retry_interval"])
                    if self.options["retry_interval_add"]:
                        self.options["retry_interval"] += self.options[
                            "retry_interval_add"
                        ]
                self.options["retries"] -= 1
                self.logger.warning(
                    f"Retrying ({self.options['retries']} attempts remaining)"
                )

    def _try(self):
        raise NotImplementedError("Subclasses should provide their own implementation")

    def _is_retry_valid(self, e):
        return True

    def _reset_poll(self):
        self.poll_complete = False
        self.poll_count = 0
        self.poll_interval_level = 0
        self.poll_interval_s = 1

    def _poll(self):
        """poll for a result in a loop"""
        while True:
            self.poll_count += 1
            self._poll_action()
            if self.poll_complete:
                break
            time.sleep(self.poll_interval_s)
            self._poll_update_interval()

    def _poll_action(self):
        """
        Poll something and process the response.
        Set `self.poll_complete = True` to break polling loop.
        """
        raise NotImplementedError("Subclasses should provide their own implementation")

    def _poll_update_interval(self):
        """update the polling interval to be used next iteration"""
        # Increase by 1 second every 3 polls
        if self.poll_count // 3 > self.poll_interval_level:
            self.poll_interval_level += 1
            self.poll_interval_s += 1
            self.logger.info(
                "Increased polling interval to %d seconds", self.poll_interval_s
            )

    def freeze(self, step):
        ui_step = {
            "name": self.task_config.name or self.name,
            "kind": "other",
            "is_required": True,
        }
        ui_step.update(self.task_config.config.get("ui_options", {}))
        task_config = {"options": self.options, "checks": self.task_config.checks or []}
        ui_step.update(
            {
                "path": step.path,
                "step_num": str(step.step_num),
                "task_class": self.task_config.class_path,
                "task_config": task_config,
                "source": step.project_config.source.frozenspec,
            }
        )
        return [ui_step]


class BaseSalesforceTask(BaseTask):
    """Base for tasks that need a Salesforce org"""

    name = "BaseSalesforceTask"
    salesforce_task = True

    def _get_client_name(self):
        try:
            app = self.project_config.keychain.get_service("connectedapp")
            return app.client_id
        except (ServiceNotValid, ServiceNotConfigured):
            return f"CumulusCI/{__version__}"

    def _run_task(self):
        raise NotImplementedError("Subclasses should provide their own implementation")

    def _update_credentials(self):
        with self.org_config.save_if_changed():
            self.org_config.refresh_oauth_token(self.project_config.keychain)



from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock

from cumulusci.core.config import OrgConfig, TaskConfig
from cumulusci.tests.util import DummyKeychain, create_project_config


def create_task(task_class, options=None, project_config=None, org_config=None):
    "Older create_task helper which does not support orginfo cache."
    if project_config is None:
        project_config = create_project_config("TestRepo", "TestOwner")
    if org_config is None:
        org_config = OrgConfig(
            {
                "instance_url": "https://test.salesforce.com",
                "id": "https://test.salesforce.com/ORG_ID/USER_ID",
                "access_token": "TOKEN",
                "org_id": "ORG_ID",
                "username": "test-cci@example.com",
            },
            "test",
            keychain=DummyKeychain(),
        )
        org_config.refresh_oauth_token = mock.Mock()
    if options is None:
        options = {}
    task_config = TaskConfig({"options": options})
    with mock.patch(
        "cumulusci.tasks.salesforce.BaseSalesforceTask._get_client_name",
        return_value="ccitests",
    ):
        return task_class(project_config, task_config, org_config)


def patch_dir(patch_path, file_path):
    directory = Path(file_path)
    directory.mkdir(parents=True, exist_ok=True)
    patch = mock.patch(patch_path, lambda: file_path)
    patch.start()
    return patch


def create_task_fixture(request):
    "create_task fixture which does support orginfo cache."
    temp_dirs = TemporaryDirectory()
    temp_root = Path(temp_dirs.name)

    to_patch = {
        "pathlib.Path.home": temp_root / "fixture_user_home",
        "pathlib.Path.cwd": temp_root / "fixture_user_home/project",
    }

    pretend_git = temp_root / "fixture_user_home/project/.git"
    pretend_git.mkdir(parents=True)
    patches = [patch_dir(p, d) for p, d in to_patch.items()]

    def fin():
        for patch in patches:
            patch.stop()

        temp_dirs.cleanup()

    request.addfinalizer(fin)

    return create_task



from cumulusci.core.config import BaseProjectConfig, TaskConfig, UniversalConfig
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.tasks.bulkdata.step import (
    BaseDmlOperation,
    DataOperationJobResult,
    DataOperationStatus,
)
from cumulusci.tests import util as cci_test_utils


def _make_task(task_class, task_config):
    task_config = TaskConfig(task_config)
    universal_config = UniversalConfig()
    project_config = BaseProjectConfig(
        universal_config,
        config={"noyaml": True, "project": {"package": {"api_version": "46.0"}}},
    )
    keychain = BaseProjectKeychain(project_config, "")
    project_config.set_keychain(keychain)
    org_config = cci_test_utils.DummyOrgConfig(
        {"instance_url": "https://example.com", "access_token": "abc123"}, "test"
    )
    return task_class(project_config, task_config, org_config)


class FakeBulkAPI:
    """Extremely simplistic mock of the bulk API

    Can be improved as needed over time.
    """

    next_job_id = 0
    next_batch_id = 0

    @classmethod
    def create_job(cls, *args, **kwargs):
        cls.next_job_id += 1
        return cls.next_job_id

    @classmethod
    def create_query_job(cls, *args, **kwargs):
        cls.next_job_id += 1
        return cls.next_job_id

    @classmethod
    def post_batch(cls, *args, **kwargs):
        cls.next_batch_id += 1
        return cls.next_batch_id

    def close_job(self, *args, **kwargs):
        pass

    def job_status(self, job_id):
        return {
            "numberBatchesCompleted": self.next_batch_id,
            "numberBatchesTotal": self.next_batch_id,
        }

    def query(self, job_id, soql):
        return self.post_batch()

    def get_query_batch_result_ids(self, batch_id, job_id):
        return range(0, 10)


class FakeBulkAPIDmlOperation(BaseDmlOperation):
    def __init__(
        self, *, context, sobject=None, operation=None, api_options=None, fields=None
    ):
        super().__init__(
            sobject=sobject,
            operation=operation,
            api_options=api_options,
            context=context,
            fields=fields,
        )
        self._results = []
        self.records = []

    def start(self):
        self.job_id = "JOB"

    def end(self):
        records_processed = len(self.results)
        self.job_result = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], records_processed, 0
        )

    def load_records(self, records):
        self.records.extend(records)

    def get_results(self):
        return iter(self.results)

    @property
    def results(self):
        return self._results

    @results.setter
    def results(self, results):
        self._results = results
        self.end()



import platform
import sys
import webbrowser
from datetime import datetime
from pathlib import Path

import click
import github3

import cumulusci
from cumulusci.core.exceptions import CumulusCIException
from cumulusci.core.github import check_github_scopes, create_gist, get_github_api

from .runtime import pass_runtime

GIST_404_ERR_MSG = """A 404 error code was returned when trying to create your gist.
Please ensure that your GitHub personal access token has the 'Create gists' scope."""


@click.group("error", short_help="Get or share information about an error")
def error():
    """
    Get or share information about an error

    If you'd like to dig into an error more yourself,
    you can get the last few lines of context about it
    from `cci error info`.

    If you'd like to submit it to a developer for conversation,
    you can use the `cci error gist` command. Just make sure
    that your GitHub access token has the 'create gist' scope.

    If you'd like to regularly see stack traces, set the `show_stacktraces`
    option to `True` in the "cli" section of `~/.cumulusci/cumulusci.yml`, or to
    see a stack-trace (and other debugging information) just once, use the `--debug`
    command line option.

    For more information on working with errors in CumulusCI visit:
    https://cumulusci.readthedocs.io/en/latest/features.html#working-with-errors
    """


CCI_LOGFILE_PATH = Path.home() / ".cumulusci" / "logs" / "cci.log"


@error.command(
    name="info",
    help="Outputs the most recent traceback (if one exists in the most recent log)",
)
def error_info():
    if not CCI_LOGFILE_PATH.is_file():
        click.echo(f"No logfile found at: {CCI_LOGFILE_PATH}")
        return

    traceback = get_traceback(CCI_LOGFILE_PATH.read_text(encoding="utf-8"))
    click.echo(traceback)


def get_traceback(log_content: str) -> str:
    """Returns the the traceback in a logfile if it exists."""
    stacktrace_start = "Traceback (most recent call last):"
    if stacktrace_start not in log_content:
        return f"\nNo stacktrace found in: {CCI_LOGFILE_PATH}\n"

    stacktrace = ""
    for i, line in enumerate(reversed(log_content.split("\n")), 1):
        stacktrace = "\n" + line + stacktrace
        if stacktrace_start in line:
            break

    return stacktrace


@error.command(name="gist", help="Creates a GitHub gist from the latest logfile")
@pass_runtime(require_project=False, require_keychain=True)
def gist(runtime):
    if CCI_LOGFILE_PATH.is_file():
        log_content = CCI_LOGFILE_PATH.read_text(encoding="utf-8")
    else:
        log_not_found_msg = """No logfile to open at path: {}
        Please ensure you're running this command from the same directory you were experiencing an issue."""
        error_msg = log_not_found_msg.format(CCI_LOGFILE_PATH)
        click.echo(error_msg)
        raise CumulusCIException(error_msg)

    last_cmd_header = "\n\n\nLast Command Run\n================================\n"
    filename = f"cci_output_{datetime.utcnow()}.txt"
    files = {
        filename: {"content": f"{get_context_info()}{last_cmd_header}{log_content}"}
    }

    try:
        gh = runtime.keychain.get_service("github")
        gist = create_gist(
            get_github_api(gh.username, gh.password or gh.token),
            "CumulusCI Error Output",
            files,
        )
    except github3.exceptions.NotFoundError as exc:
        scope_warning = check_github_scopes(exc)
        raise CumulusCIException(GIST_404_ERR_MSG + f"\n\n{scope_warning}")
    except Exception as e:
        raise CumulusCIException(
            f"An error occurred attempting to create your gist:\n{e}"
        )
    else:
        click.echo(f"Gist created: {gist.html_url}")
        webbrowser.open(gist.html_url)


def get_context_info():
    host_info = platform.uname()

    info = []
    info.append(f"CumulusCI version: {cumulusci.__version__}")
    info.append(f"Python version: {sys.version.split()[0]} ({sys.executable})")
    info.append(f"Environment Info: {host_info.system} / {host_info.machine}")
    return "\n".join(info)



"""
Tests for the robot task, specifically for options related to running tests in parallel
"""

import sys
import unittest
from pathlib import Path
from unittest import mock

from cumulusci.tasks.robotframework import Robot
from cumulusci.tasks.salesforce.tests.util import create_task


class TestRobotParallel(unittest.TestCase):
    """Tests for the robot task when running tests with pabot"""

    @mock.patch("cumulusci.tasks.robotframework.robotframework.robot_run")
    @mock.patch("cumulusci.tasks.robotframework.robotframework.subprocess.run")
    def test_process_arg_gt_zero(self, mock_subprocess_run, mock_robot_run):
        """Verify that setting the process option to a number > 1 runs pabot"""
        mock_subprocess_run.return_value = mock.Mock(returncode=0)
        task = create_task(Robot, {"suites": "tests", "processes": "2"})
        task()
        outputdir = str(Path(".").resolve())
        expected_cmd = [
            sys.executable,
            "-m",
            "pabot.pabot",
            "--pabotlib",
            "--processes",
            "2",
            "--pythonpath",
            task.project_config.repo_root,
            "--variable",
            "org:test",
            "--outputdir",
            outputdir,
            "--tagstatexclude",
            "cci_metric_elapsed_time",
            "--tagstatexclude",
            "cci_metric",
            "tests",
        ]
        mock_robot_run.assert_not_called()
        mock_subprocess_run.assert_called_once_with(expected_cmd)

    @mock.patch("cumulusci.tasks.robotframework.robotframework.robot_run")
    @mock.patch("cumulusci.tasks.robotframework.robotframework.subprocess.run")
    def test_ordering_arg_processes_gt_1(self, mock_subprocess_run, mock_robot_run):
        mock_subprocess_run.return_value = mock.Mock(returncode=0)
        task = create_task(
            Robot, {"suites": "tests", "processes": "2", "ordering": "robot/order.txt"}
        )
        task()
        outputdir = str(Path(".").resolve())
        expected_cmd = [
            sys.executable,
            "-m",
            "pabot.pabot",
            "--pabotlib",
            "--processes",
            "2",
            # note: we are specifically testing that this comes before --pythonpath
            "--ordering",
            "robot/order.txt",
            "--pythonpath",
            task.project_config.repo_root,
            "--variable",
            "org:test",
            "--outputdir",
            outputdir,
            "--tagstatexclude",
            "cci_metric_elapsed_time",
            "--tagstatexclude",
            "cci_metric",
            "tests",
        ]
        mock_robot_run.assert_not_called()
        mock_subprocess_run.assert_called_once_with(expected_cmd)

    @mock.patch("cumulusci.tasks.robotframework.robotframework.robot_run")
    @mock.patch("cumulusci.tasks.robotframework.robotframework.subprocess.run")
    def test_testlevelsplit_arg_processes_gt_1(
        self, mock_subprocess_run, mock_robot_run
    ):
        mock_subprocess_run.return_value = mock.Mock(returncode=0)
        task = create_task(
            Robot, {"suites": "tests", "processes": "2", "testlevelsplit": "true"}
        )
        task()
        outputdir = str(Path(".").resolve())
        expected_cmd = [
            sys.executable,
            "-m",
            "pabot.pabot",
            "--pabotlib",
            "--processes",
            "2",
            # note: we are specifically testing that this comes before --pythonpath
            # and that the option isn't followed by a value
            "--testlevelsplit",
            "--pythonpath",
            task.project_config.repo_root,
            "--variable",
            "org:test",
            "--outputdir",
            outputdir,
            "--tagstatexclude",
            "cci_metric_elapsed_time",
            "--tagstatexclude",
            "cci_metric",
            "tests",
        ]
        mock_robot_run.assert_not_called()
        mock_subprocess_run.assert_called_once_with(expected_cmd)

    @mock.patch("cumulusci.tasks.robotframework.robotframework.robot_run")
    @mock.patch("cumulusci.tasks.robotframework.robotframework.subprocess.run")
    def test_pabot_args_process_eq_zero(self, mock_subprocess_run, mock_robot_run):
        """Verifies that pabot-specific options are not passed on when processes=1"""
        mock_robot_run.return_value = 0
        task = create_task(
            Robot,
            {
                "suites": "tests",
                "processes": 1,
                "testlevelsplit": "true",
                "ordering": "./robot/order.txt",
            },
        )
        task()
        mock_subprocess_run.assert_not_called()
        outputdir = str(Path(".").resolve())
        mock_robot_run.assert_called_once_with(
            "tests",
            listener=[],
            outputdir=outputdir,
            variable=["org:test"],
            tagstatexclude=["cci_metric_elapsed_time", "cci_metric"],
            stdout=sys.stdout,
            stderr=sys.stderr,
        )



import re
import typing as T
from io import StringIO
from logging import getLogger
from pathlib import Path

import yaml
from yaml.error import MarkedYAMLError

from cumulusci.core.exceptions import YAMLParseException
from cumulusci.utils.fileutils import FSResource, load_from_source

NBSP = "\u00A0"

pattern = re.compile(r"^\s*[\u00A0]+\s*", re.MULTILINE)

logger = getLogger(__name__)


def _replace_nbsp(origdata, filename):
    """Replace nbsp characters in leading whitespace in a YAML file."""
    counter = 0

    def _replacer_func(matchobj):
        nonlocal counter
        counter += 1
        string = matchobj.group(0)
        rc = string.replace(NBSP, " ")
        return rc

    data = pattern.sub(_replacer_func, origdata)

    if counter:
        plural = "s were" if counter > 1 else " was"
        logger.warning(
            f"Note: {counter} lines with non-breaking space character{plural} detected in {filename}.\n"
            "Perhaps you cut and pasted from a Web page?\n"
            "Future versions of CumulusCI may disallow these characters.\n"
        )
    return data


def load_yaml_data(
    source: T.Union[str, T.IO[T.Text], Path, FSResource], context: str = None
):
    """Load a file, convert NBSP->space and parse it in YAML.

    Raises YAMLParseException with a nicely formatted error message
    if an error occurs while parsing.

    If you use this method directly (or, heaven forbid, yaml.safe_load)
    consider making a CCIModel subclass instead.
    """
    with load_from_source(source) as (f_config, filename):
        if filename == "<stream>":
            filename = "a yaml file"
        context = context or filename
        data = _replace_nbsp(f_config.read(), context)
        try:
            rc = yaml.safe_load(StringIO(data))
        except MarkedYAMLError as e:
            line_num = e.problem_mark.line + 1
            column_num = e.problem_mark.column
            message = (
                f"An error occurred parsing yaml file at line {line_num}, column {column_num}.\n"
                + f"{filename}:{line_num}:{column_num}\n"
                + f"Error message: {e.problem}."
            )
            raise YAMLParseException(message)
        except Exception as e:
            message = f"An error occurred parsing {filename}.\nError message: {e}"
            raise YAMLParseException(message)

        return rc
